## 1.5. 獲取URL

對於很多現代應用來説，訪問互聯網上的信息和訪問本地文件繫統一樣重要。Go語言在net這個強大package的幫助下提供了一繫列的package來做這件事情，使用這些包可以更簡單地用網絡收發信息，還可以建立更底層的網絡連接，編寫服務器程序。在這些情景下，Go語言原生的併發特性（在第八章中會介紹）顯得尤其好用。

爲了最簡單地展示基於HTTP獲取信息的方式，下面給出一個示例程序fetch，這個程序將獲取對應的url，併將其源文本打印出來；這個例子的靈感來源於curl工具（譯註：unix下的一個用來發http請求的工具，具體可以man curl）。當然，curl提供的功能更爲複雜豐富，這里隻編寫最簡單的樣例。這個樣例之後還會多次被用到。

<u><i>gopl.io/ch1/fetch</i></u>
```go
// Fetch prints the content found at a URL.
package main

import (
	"fmt"
	"io/ioutil"
	"net/http"
	"os"
)

func main() {
	for _, url := range os.Args[1:] {
		resp, err := http.Get(url)
		if err != nil {
			fmt.Fprintf(os.Stderr, "fetch: %v\n", err)
			os.Exit(1)
		}
		b, err := ioutil.ReadAll(resp.Body)
		resp.Body.Close()
		if err != nil {
			fmt.Fprintf(os.Stderr, "fetch: reading %s: %v\n", url, err)
			os.Exit(1)
		}
		fmt.Printf("%s", b)
	}
}
```

這個程序從兩個package中導入了函數，net/http和io/ioutil包，http.Get函數是創建HTTP請求的函數，如果獲取過程沒有出錯，那麽會在resp這個結構體中得到訪問的請求結果。resp的Body字段包括一個可讀的服務器響應流。ioutil.ReadAll函數從response中讀取到全部內容；將其結果保存在變量b中。resp.Body.Close關閉resp的Body流，防止資源洩露，Printf函數會將結果b寫出到標準輸出流中。

```
$ go build gopl.io/ch1/fetch
$ ./fetch http://gopl.io
<html>
<head>
<title>The Go Programming Language</title>title>
...
```

HTTP請求如果失敗了的話，會得到下面這樣的結果：

```
$ ./fetch http://bad.gopl.io
fetch: Get http://bad.gopl.io: dial tcp: lookup bad.gopl.io: no such host
```

譯註：在大天朝的網絡環境下很容易重現這種錯誤，下面是Windows下運行得到的錯誤信息：

```
$ go run main.go http://gopl.io
fetch: Get http://gopl.io: dial tcp: lookup gopl.io: getaddrinfow: No such host is known.
```

無論哪種失敗原因，我們的程序都用了os.Exit函數來終止進程，併且返迴一個status錯誤碼，其值爲1。

**練習 1.7：** 函數調用io.Copy(dst, src)會從src中讀取內容，併將讀到的結果寫入到dst中，使用這個函數替代掉例子中的ioutil.ReadAll來拷貝響應結構體到os.Stdout，避免申請一個緩衝區（例子中的b）來存儲。記得處理io.Copy返迴結果中的錯誤。

**練習 1.8：** 脩改fetch這個范例，如果輸入的url參數沒有 `http://` 前綴的話，爲這個url加上該前綴。你可能會用到strings.HasPrefix這個函數。

**練習 1.9：** 脩改fetch打印出HTTP協議的狀態碼，可以從resp.Status變量得到該狀態碼。
